/*
K1n9_Duk3's IMF to WAV converter - Converts IMF files to WAV.
Copyright (C) 2013 K1n9_Duk3

Based on Wolf4SDL by Moritz "Ripper" Kroll (http://www.chaos-software.de.vu)

The OPL emulator (fmopl.cpp, fmopl.h) is used under the terms of the
MAME license (see license-mame.txt for more details). 

Redistributions of this program may not be sold, nor may they be used 
in a commercial product or activity, unless a different OPL emulator 
is used.

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "fmopl.h"

#define YM3812_RATE 3579545
#define DEF_FREQ 44100

//just a quick hack, but it works...
#pragma pack(2)		//this gave me a headache...
struct waveheader{
	UINT32 rID, rSize;
	UINT32 wID;
	UINT32 fID, fSize;
	UINT16 fFormat, fChannels;
	UINT32 fHertz, fBytesPerSec;
	UINT16 fBlockAlign, fBits, fSpecific;
	UINT32 dID, dSize;
};
#pragma pack()

waveheader head = 
{
	0x46464952,		//rID = "RIFF"
	0,				//rSize (dummy value)
	0x45564157,		//wID = "WAVE"
	0x20746D66,		//fID = "fmt "
	18,				//fSize
	1,				//fFormat
	1,				//fChannels
	DEF_FREQ,		//fHertz
	0,				//fBytesPerSec (dummy value)
	4,				//fBlockAlign
	OPL_SAMPLE_BITS,//fBits
	0,				//fSpecific
	0x61746164,		//dID = "data"
	0				//dSize (dummy value)
};

int IMF_IsChunk(FILE* in)
{
	UINT16 chunksize, buff, i=42;
	UINT32 sum1=0, sum2=0;

	if(!in)
		return 0;

	if (!feof(in))
	{
		fread(&chunksize, sizeof(chunksize), 1, in);
		while(!feof(in) && i)
		{
			fread(&buff, sizeof(buff), 1, in);
			sum1 += buff;
			fread(&buff, sizeof(buff), 1, in);
			sum2 += buff;
			i--;
		}
		fseek(in, 0, SEEK_SET);
		return (sum1 > sum2);
	}
	return 0;
}

int main(int argc, char *argv[])
{
    FILE *in, *out;
    INT16 *buffer;
	char *wavefile;
    unsigned int imf_rate=560, wav_rate=DEF_FREQ, samples_per_tick;

	printf("K1n9_Duk3's IMF to WAV converter v1.0\n");
	printf("=====================================\n\n");
	if(argc < 2)
    {
        printf("Usage: IMF2WAV <imffile> [wavefile [imfrate [hertz]]]\n");
        return 0;
	} else {
		if(argc==2)
		{
			wavefile = (char*)malloc(strlen(argv[1])+5);
			strcpy(wavefile, argv[1]);
			strcat(wavefile, ".wav");
		} else {
			wavefile = argv[2];
		}

		if(argc > 3) {
			imf_rate = atoi(argv[3]);
			if (argc > 4)
				wav_rate = atoi(argv[4]);
		}
	}
	samples_per_tick=wav_rate/imf_rate;
    if(YM3812Init(1, YM3812_RATE, wav_rate))
    {
        printf("Unable to create virtual OPL!!\n");
        return 1;
    }
    printf("OPL created.\n");
    for(int i=1;i<0xf6;i++)
        YM3812Write(0,i,0);

    YM3812Write(0,1,0x20); // Set WSE=1
    //YM3812Write(0,8,0); // Set CSM=0 & SEL=0           // already set in for statement

    in = fopen(argv[1], "rb");
    if(in)
    {
        printf("%s opened for input.\n", argv[1]);
        out = fopen(wavefile, "wb");
        if(out)
        {
            UINT16 imfdelay, imfcommand;
			UINT32 size, cnt=0, imfsize=0xFFFFFFFF;
            printf("%s opened for output.\n", wavefile);
            buffer = (INT16 *)malloc(samples_per_tick*sizeof(INT16));
            if (buffer)
            {
                printf("buffer allocated.\n");
				if (IMF_IsChunk(in))
				{
					printf("IMF file is a ripped AudioT chunk.\n");
					fread(&imfcommand, sizeof(INT16), 1, in);
	                printf("IMF size is %u Bytes.\n", imfcommand);
					imfsize = imfcommand >> 2;
				} else {
					printf("IMF size is not set.\n");
				}
                printf("IMF rate is %u Hz.\n", imf_rate);
				printf("Converting IMF data to PCM data\n");

				//write dummy wave header:
				fwrite(&head, sizeof(head), 1, out);	
				while(!feof(in) && imfsize)
                {
					imfsize--;
                    fread(&imfcommand, sizeof(INT16), 1, in);
                    fread(&imfdelay, sizeof(INT16), 1, in);
                    YM3812Write(0, imfcommand & 0xFF, (imfcommand >> 8) & 0xFF);
                    while(imfdelay--)
                    {
                        YM3812UpdateOne(0, buffer, samples_per_tick);
                        fwrite(buffer, sizeof(INT16), samples_per_tick, out);
                        fflush(out);
						if (feof(in)) break;
                    }
					if(!(cnt++ & 0xff))
					{
						printf(".");
						fflush(stdout);
					}
                }
				printf(" done!\n");
            } else {
                printf("ERROR: out of memory\n");
            }
			size = ftell(out);

			//fill header with correct values:
			head.dSize = size-sizeof(head);
			head.rSize = size-8;
			head.fHertz = wav_rate;
			head.fBlockAlign = head.fChannels*(head.fBits/8);
			head.fBytesPerSec = head.fBlockAlign*wav_rate;

			//write real wave header:
			fseek(out, 0, SEEK_SET);
			fwrite(&head, sizeof(head), 1, out);	
			fclose(out);
        } else {
            printf("ERROR: could not write %s\n", wavefile);
        }
        fclose(in);
    } else {
        printf("ERROR: could not read %s\n", argv[1]);
    }

    YM3812Shutdown();
    return 0;
}
